from django.conf import settings
from django.core.checks import Error, Tags, register


@register(Tags.compatibility)
def dependencies_check(app_configs, **kwargs):
    errors = []

    if "django_otp" not in settings.INSTALLED_APPS:
        errors.append(
            Error(
                "zweifach app requires django-otp to be installed.",
                id="zweifach.E001",
            )
        )

    if "django_otp.middleware.OTPMiddleware" not in settings.MIDDLEWARE:
        errors.append(
            Error(
                "zweifach app requires OTPMiddleware to be enabled.",
                id="zweifach.E002",
                hint="OTPMiddleware is included in django-otp package",
            )
        )
    return errors



from django.urls import path
from django.views.decorators.cache import never_cache

from zweifach import views


urlpatterns = [
    path('verify/', never_cache(views.verify), name='zweifach_verify'),
    path('setup/', never_cache(views.setup), name='zweifach_setup'),
]



from django_otp import login, match_token, verify_token

from django import forms


class VerifyForm(forms.Form):
    code = forms.CharField()

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        super().__init__(*args, **kwargs)

    def clean_code(self):
        device = match_token(self.request.user, self.cleaned_data["code"])
        if not device:
            raise forms.ValidationError("code did not validate")
        else:
            login(self.request, device)


class SetupForm(forms.Form):
    code = forms.CharField()

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop("request")
        self.device = kwargs.pop("device")
        super().__init__(*args, **kwargs)

    def clean_code(self):
        device = verify_token(
            self.request.user,
            self.device.persistent_id,
            self.cleaned_data["code"],
        )
        if not device == self.device:
            raise forms.ValidationError("code did not validate")
        else:
            device.confirmed = True
            device.save(update_fields=("confirmed",))
            login(self.request, device)



from django_otp import user_has_device

from django.conf import settings
from django.shortcuts import redirect
from django.template.defaultfilters import urlencode
from django.urls import reverse


def two_factor_auth_required(user):
    """Check if 'user' must enable 2FA."""
    for check in getattr(settings, 'ZWEIFACH_AUTH_REQUIRED', []):
        if check(user):
            return True
    return False


def two_factor_auth_configured(user):
    """Check if 'user' has 2FA enabled."""
    if user.is_authenticated:
        return user_has_device(user)
    return False


def excluded_urls():
    """Return a list of URLs which are accessible without verifying the second factor."""
    default_excludes = [reverse('zweifach_verify'), reverse('zweifach_setup')]
    if 'django.contrib.admin' in settings.INSTALLED_APPS:
        default_excludes.append(reverse('admin:login'))
    if settings.LOGIN_URL:
        default_excludes.append(settings.LOGIN_URL)

    return getattr(settings, 'ZWEIFACH_URL_EXCLUDES', []) + default_excludes


class ZweifachMiddleware:
    """
    - Enforce 2FA, if user has it configured.
    - Enforce 2FA-Setup, if user is required to use it.
    """

    def __init__(self, get_response):
        self.get_response = get_response
        self.excluded_urls = excluded_urls()

    def __call__(self, request):

        if request.path not in self.excluded_urls:

            if two_factor_auth_required(
                request.user
            ) and not two_factor_auth_configured(request.user):
                return redirect(
                    reverse('zweifach_setup')
                    + f'?next={urlencode(request.get_full_path())}'
                )

            if (
                two_factor_auth_configured(request.user)
                and not request.user.is_verified()
            ):
                return redirect(
                    reverse('zweifach_verify')
                    + f'?next={urlencode(request.get_full_path())}'
                )

        return self.get_response(request)



import io

import qrcode
import qrcode.image.svg
from django_otp import user_has_device
from django_otp.plugins.otp_totp.models import TOTPDevice

from django.shortcuts import redirect, render

from zweifach.forms import SetupForm, VerifyForm


def verify(request):
    """Verify second auth factor."""
    if request.method == "POST":
        form = VerifyForm(request.POST, request=request)
        if form.is_valid():
            return redirect(request.GET.get("next", "/"))
    else:
        form = VerifyForm(request=request)
    return render(request, "zweifach/verify.html", {'form': form})


def setup(request):
    """Setup two factor auth for user."""
    if user_has_device(request.user):
        raise NotImplementedError("already configured!")

    # create TOTP device
    device = TOTPDevice.objects.get_or_create(
        user=request.user,
        confirmed=False,
        defaults={"name": f"{request.user.username}'s device"},
    )[0]

    # display QR-Code (provided as base64 string, so we don't need any url configuration)
    qr_code = io.BytesIO()
    img = qrcode.make(
        device.config_url, image_factory=qrcode.image.svg.SvgImage
    )
    img.save(qr_code)

    # display form
    if request.method == "POST":
        form = SetupForm(request.POST, request=request, device=device)
        # setup is done, if the entered code is valid
        if form.is_valid():
            return redirect(request.GET.get("next", "/"))
    else:
        form = SetupForm(request=request, device=device)

    return render(
        request,
        "zweifach/setup.html",
        {
            'form': form,
            'qrcode': qr_code.getvalue().decode("utf-8"),
            'device': device,
        },
    )
